// This file is part of Doodah Band.
//
// Doodah Band is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//
// Doodah Band is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with Doodah Band. If not, see <https://www.gnu.org/licenses/>.

// Audio subsystem that loads and plays an OGG compressed audio file.
// It also analyses the audio being played to feed visualisations.
// Note that this module does not support being initialised more than once.

// Number of samples used for audio analysis.
// This needs to be a power of two because:
// - FFTs tend to perform well at these sizes
// - some OpenGL implementations don't like non-power-of-two textures
//   (FFT output will be fed to OpenGL via a texture)
#define ANALYSIS_WINDOW_SAMPLES 512

// Used in audio callback.
// Only touch these values while using `SDL_LockAudioDevice()`.
	// OGG decoder for currently open audio file (if any).
	OggVorbis_File audio_file;
	// `true` when a file is open implying that `audio_file` is valid.
	bool audio_file_open = false;
	// Current audio format and sample rate converter, if any.
	SDL_AudioStream *audio_converter = NULL;
	// Configuration of `audio_converter`.
	int audio_converter_channels = 0;
	int audio_converter_rate = 0;
	// Sample buffer used to hold rendered audio for analyises.
	// Length is always `ANALYSIS_WINDOW_SAMPLES+audio_format.samples` to ensure
	// that there will be enough data to analyse for the entire time the sample
	// buffer is being played.
	int16_t *audio_analysis_buffer = NULL;
	// Index of start of current window in analysis buffer.
	size_t audio_analysis_offset = 0;

// Current audio device.
SDL_AudioDeviceID audio_device = 0;
// Parameters of `audio_device`.
SDL_AudioSpec audio_format;
// Copy of current window in analysis buffer.
// Copying allows slow analyses to run without holding up the audio thread.
int16_t audio_analysis_copy[ANALYSIS_WINDOW_SAMPLES];
// State used by FFT analysis.
kiss_fftr_cfg audio_fft = NULL;

// Internal function (requires audio device to be locked).
// Stop playing current file.
void audio_locked_stop(void) {
	if (audio_file_open) ov_clear(&audio_file);
	audio_file_open = false;

	SDL_FreeAudioStream(audio_converter);
	audio_converter = NULL;
}

// Internal function.
// Fill an audio output buffer.
void audio_callback(void *const userdata, uint8_t *const stream, const int length) {
	(void)userdata;

	// Decode and resample audio until out of file.
	uint8_t buffer[1024];
	int offset = 0;
	while (offset < length) {
		// Decode more audio when converter buffer is empty.
		while (audio_converter == NULL || SDL_AudioStreamAvailable(audio_converter) == 0) {
			// Stop if no file to read from.
			if (!audio_file_open) break;

			// Get information about next section of file.
			const vorbis_info *const info = ov_info(&audio_file, -1);
			if (info == NULL) {
				// Read error.
				audio_locked_stop();
				break;
			}

			// Change converter setup to match section.
			if (audio_converter == NULL || info->channels != audio_converter_channels || info->rate != audio_converter_rate) {
				if (audio_converter != NULL) SDL_FreeAudioStream(audio_converter);

				audio_converter = SDL_NewAudioStream(
					AUDIO_S16LSB, info->channels, info->rate,
					audio_format.format, audio_format.channels, audio_format.freq);
				if (audio_converter == NULL) {
					// Failed to create converter.
					audio_locked_stop();
					break;
				}

				audio_converter_channels = info->channels;
				audio_converter_rate = info->rate;
			}

			// Read data from file.
			int next_section = 0;
			const long bytes = ov_read(&audio_file, (char*)buffer, sizeof(buffer), 0, 2, 1, &next_section);
			if (bytes <= 0) {
				// End of file or read error, stop playing.
				audio_locked_stop();
				break;
			}

			// Feed converter.
			if (SDL_AudioStreamPut(audio_converter, buffer, bytes) != 0) abort();
		}
		
		// Read data from converter.
		if (audio_converter == NULL) break;
		const int bytes = SDL_AudioStreamGet(audio_converter, &stream[offset], length-offset);
		if (bytes == 0) break;
		if (bytes <= 0) abort();
		offset += bytes;
	}

	// Fill remaining buffer with silence.
	while (offset < length) {
		stream[offset] = 0;
		offset ++;
	}

	// Move end of analysis buffer to beginning.
	const size_t frames = length/sizeof(int16_t)/audio_format.channels;
	const size_t remaining = audio_format.samples+ANALYSIS_WINDOW_SAMPLES-frames;
	if (frames < (size_t)audio_format.samples+ANALYSIS_WINDOW_SAMPLES) {
		const int16_t *source = &audio_analysis_buffer[frames];
		int16_t *destination = &audio_analysis_buffer[0];
		size_t counter;
		for (counter = 0; counter < remaining; counter ++) {
			*destination = *source;
			source ++;
			destination ++;
		}
	}

	// Copy one channel of audio to end of analysis buffer.
	const int16_t *source = (const int16_t *)stream;
	int16_t *destination = &audio_analysis_buffer[remaining];
	size_t counter;
	for (counter = 0; counter < frames; counter ++) {
		*destination = *source;
		source += audio_format.channels;
		destination ++;
	}

	// Reset analysis position.
	audio_analysis_offset = 0;
}

// Initialise audio subsystem into idle (no file playing).
void audio_init(void) {
	// Open output device.
	const SDL_AudioSpec desired_format = {
		.freq = 48000,
		.format = AUDIO_S16LSB,
		.channels = 2,
		.silence = 0,
		.samples = 2048,
		.size = 0,
		.callback = audio_callback,
		.userdata = NULL,
	};
	audio_device = SDL_OpenAudioDevice(NULL, 0, &desired_format, &audio_format, SDL_AUDIO_ALLOW_FREQUENCY_CHANGE | SDL_AUDIO_ALLOW_CHANNELS_CHANGE | SDL_AUDIO_ALLOW_SAMPLES_CHANGE);
	if (audio_device <= 0) abort_sdl("failed to open audio output");

	// Allocate buffers.
	audio_analysis_buffer = malloc(sizeof(audio_analysis_buffer[0])*(ANALYSIS_WINDOW_SAMPLES+audio_format.samples));
	if (audio_analysis_buffer == NULL) abort();

	// Initialise FFT.
	audio_fft = kiss_fftr_alloc(ANALYSIS_WINDOW_SAMPLES, 0, NULL, NULL);
	if (audio_fft == NULL) abort();

	// Start playing silence.
	SDL_PauseAudioDevice(audio_device, 0);
}

// Start playing given file if no file is currently playing.
// Returns `true` if nothing was playing (signal to remove file from playlist).
bool audio_queue(const char *path) {
	if (path == NULL) abort();

	bool consumed = false;
	SDL_LockAudioDevice(audio_device);
	if (!audio_file_open) {
		consumed = true;
		if (ov_fopen(path, &audio_file) != 0) {
			fprintf(stderr, "ERROR: failed to open track \"%s\"\n", path);
		} else {
			audio_file_open = true;
		}
	}
	SDL_UnlockAudioDevice(audio_device);

	return consumed;
}

// Pause or unpause anything that is playing.
void audio_pause(const bool pause_audio) {
	if (pause_audio) {
		SDL_PauseAudioDevice(audio_device, 1);
	} else {
		SDL_PauseAudioDevice(audio_device, 0);
	}
}

// Stop playing current file.
void audio_stop(void) {
	SDL_LockAudioDevice(audio_device);
	audio_locked_stop();
	SDL_UnlockAudioDevice(audio_device);
}

// Seek forward or back by given duration.
void audio_skip(const float delta_seconds) {
	SDL_LockAudioDevice(audio_device);
	double time = ov_time_tell(&audio_file)+delta_seconds;
	if (time < 0.0) time = 0.0;
	ov_time_seek_lap(&audio_file, time);
	SDL_UnlockAudioDevice(audio_device);
}

// Move analysis window by given time step, and return results of analysis.
// `amplitude` is used to keep a running average of peak amplitude (0-1).
// `fft_out` is used to return ANALYSIS_WINDOW_SAMPLES amplitudes of different
// frequency ranges.
void audio_analyse(const uint64_t delta_ms, float *amplitude, uint8_t *fft_out) {
	// Copy chunk of audio to analyse when we're not holding up the audio callback.
	SDL_LockAudioDevice(audio_device);
	{
		audio_analysis_offset += delta_ms*audio_format.freq/1000;
		if (audio_analysis_offset > audio_format.samples) audio_analysis_offset = audio_format.samples;
		memcpy(audio_analysis_copy, &audio_analysis_buffer[audio_analysis_offset], sizeof(audio_analysis_copy));
	}
	SDL_UnlockAudioDevice(audio_device);

	// Calculate peak amplitude.
	size_t index;
	int16_t peak = 0;
	for (index = 0; index < ANALYSIS_WINDOW_SAMPLES; index ++) {
		const int16_t sample = abs(audio_analysis_copy[index]);
		if (sample > peak) peak = sample;
	}
	*amplitude = *amplitude*0.8 + (float)peak/32767.0*0.2;

	// Perform FFT and convert complex results to amplitudes.
	kiss_fft_cpx fft_output[ANALYSIS_WINDOW_SAMPLES];
	kiss_fftr(audio_fft, audio_analysis_copy, fft_output);
	for (index = 0; index < ANALYSIS_WINDOW_SAMPLES/2; index ++) {
		fft_out[index] = (uint8_t)(sqrtf(
			(float)fft_output[index].r*(float)fft_output[index].r/32768.0/32768.0 +
			(float)fft_output[index].i*(float)fft_output[index].i/32768.0/32768.0)*127.0);
		fft_out[ANALYSIS_WINDOW_SAMPLES-1-index] = fft_out[index];
	}
}
