// This file is part of Doodah Band.
//
// Doodah Band is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//
// Doodah Band is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with Doodah Band. If not, see <https://www.gnu.org/licenses/>.

// Crude OpenGL binding, loader, and some support functions.

// Macro that generates bindings to OpenGL functions.
// Expects a macro `OPENGL_FUNCTION(return, name, arguments)` to be defined.
// This will be called for every OpenGL function the binding knows.
#define GENERATE_OPENGL_BINDING() \
	OPENGL_FUNCTION(void, glActiveTexture, (GLenum texture)) \
	OPENGL_FUNCTION(void, glAttachShader, (GLuint program, GLuint shader)) \
	OPENGL_FUNCTION(void, glBindAttribLocation, (GLuint program, GLuint index, const GLchar *name)) \
	OPENGL_FUNCTION(void, glBindTexture, (GLenum target, GLuint texture)) \
	OPENGL_FUNCTION(void, glClear, (GLbitfield mask)) \
	OPENGL_FUNCTION(GLuint, glCreateProgram, (void)) \
	OPENGL_FUNCTION(GLuint, glCreateShader, (GLenum type)) \
	OPENGL_FUNCTION(void, glCompileShader, (GLuint shader)) \
	OPENGL_FUNCTION(void, glCopyTexImage2D, (GLenum target, GLint level, GLenum internalformat, GLint x, GLint y, GLsizei width, GLsizei height, GLint border)) \
	OPENGL_FUNCTION(void, glDeleteProgram, (GLuint program)) \
	OPENGL_FUNCTION(void, glDeleteShader, (GLuint shader)) \
	OPENGL_FUNCTION(void, glDrawArrays, (GLenum mode, GLint first, GLsizei count)) \
	OPENGL_FUNCTION(void, glEnableVertexAttribArray, (GLuint index)) \
	OPENGL_FUNCTION(void, glFinish, (void)) \
	OPENGL_FUNCTION(void, glGenTextures, (GLsizei n, GLuint *textures)) \
	OPENGL_FUNCTION(void, glGenerateMipmap, (GLenum target)) \
	OPENGL_FUNCTION(GLenum, glGetError, (void)) \
	OPENGL_FUNCTION(void, glGetProgramiv, (GLuint program, GLenum pname, GLint *params)) \
	OPENGL_FUNCTION(void, glGetProgramInfoLog, (GLuint program, GLsizei bufSize, GLsizei *length, GLchar *infoLog)) \
	OPENGL_FUNCTION(void, glGetShaderiv, (GLuint shader, GLenum pname, GLint *params)) \
	OPENGL_FUNCTION(void, glGetShaderInfoLog, (GLuint shader, GLsizei bufSize, GLsizei *length, GLchar *infoLog)) \
	OPENGL_FUNCTION(GLint, glGetUniformLocation, (GLuint program, const GLchar *name)) \
	OPENGL_FUNCTION(void, glLinkProgram, (GLuint program)) \
	OPENGL_FUNCTION(void, glShaderSource, (GLuint shader, GLsizei count, const GLchar *const *string, const GLint *length)) \
	OPENGL_FUNCTION(void, glTexImage2D, (GLenum target, GLint level, GLint internalformat, GLsizei width, GLsizei height, GLint border, GLenum format, GLenum type, const void *pixels)) \
	OPENGL_FUNCTION(void, glTexParameteri, (GLenum target, GLenum pname, GLint param)) \
	OPENGL_FUNCTION(void, glUniform1f, (GLint location, GLfloat v0)) \
	OPENGL_FUNCTION(void, glUniform2f, (GLint location, GLfloat v0, GLfloat v1)) \
	OPENGL_FUNCTION(void, glUniform1i, (GLint location, GLint v0)) \
	OPENGL_FUNCTION(void, glUseProgram, (GLuint program)) \
	OPENGL_FUNCTION(void, glVertexAttribPointer, (GLuint index, GLint size, GLenum type, GLboolean normalized, GLsizei stride, const void *pointer)) \
	OPENGL_FUNCTION(void, glViewport, (GLint x, GLint y, GLsizei width, GLsizei height))

// Constants.
#define GL_COLOR_BUFFER_BIT 0x00004000
#define GL_COMPILE_STATUS 0x8B81
#define GL_FALSE 0
#define GL_FLOAT 0x1406
#define GL_FRAGMENT_SHADER 0x8B30
#define GL_LINEAR 0x2601
#define GL_LINEAR_MIPMAP_LINEAR 0x2703
#define GL_LINK_STATUS 0x8B82
#define GL_LUMINANCE 0x1909
#define GL_REPEAT 0x2901
#define GL_RGB 0x1907
#define GL_RGBA 0x1908
#define GL_TEXTURE0 0x84C0
#define GL_TEXTURE1 0x84C1
#define GL_TEXTURE2 0x84C2
#define GL_TEXTURE3 0x84C3
#define GL_TEXTURE_2D 0x0DE1
#define GL_TEXTURE_MAG_FILTER 0x2800
#define GL_TEXTURE_MIN_FILTER 0x2801
#define GL_TEXTURE_WRAP_S 0x2802
#define GL_TEXTURE_WRAP_T 0x2803
#define GL_TRIANGLE_STRIP 0x0005
#define GL_TRUE 1
#define GL_UNSIGNED_BYTE 0x1401
#define GL_VERTEX_SHADER 0x8B31

// Types.
typedef unsigned int GLbitfield;
typedef unsigned char GLboolean;
typedef char GLchar;
typedef unsigned int GLenum;
typedef float GLfloat;
typedef int GLint;
typedef int GLsizei;
typedef unsigned int GLuint;
typedef void GLvoid;

// Function pointers.
#define OPENGL_FUNCTION(RETURN, NAME, ARGUMENTS) RETURN (*NAME) ARGUMENTS = NULL;
GENERATE_OPENGL_BINDING()
#undef OPENGL_FUNCTION

// Loads OpenGL functions into the global function pointers.
// Note that the pointers are specific to the current GL context on some
// operating systems, so this will only work with a single context.
void load_gl_functions(const SDL_GLContext *const context) {
	(void)context; // require a context to ensure this is called after one has been setup.
	#define OPENGL_FUNCTION(RETURN, NAME, ARGUMENTS) \
		NAME = SDL_GL_GetProcAddress(#NAME); \
		if (NAME == NULL) abort();
	GENERATE_OPENGL_BINDING()
	#undef OPENGL_FUNCTION
}

// Compile a vertex or fragment shader and return it's ID.
// Returns zero on error.
GLuint compile_shader_stage(const char *const source, const size_t length, GLenum stage) {
	if (stage != GL_VERTEX_SHADER && stage != GL_FRAGMENT_SHADER) abort();
	const GLuint shader = glCreateShader(stage);
	if (shader == 0) {
		fprintf(stderr, "ERROR: failed to allocate new shader\n");
		return 0;
	}
	if (shader == 0) abort();
	GLint gl_length = length;
	glShaderSource(shader, 1, &source, &gl_length);
	glCompileShader(shader);

	GLint succeeded = GL_FALSE;
	glGetShaderiv(shader, GL_COMPILE_STATUS, &succeeded);
	if (succeeded != GL_TRUE) {
		char buffer[1024];
		GLsizei length = 0;
		glGetShaderInfoLog(shader, sizeof(buffer), &length, buffer);
		fprintf(stderr, "ERROR: failed to compile shader\n%.*s", length, buffer);
		return 0;
	}
	return shader;
}

// Link a shader and return it's ID.
// Returns zero on error.
GLuint link_shader_stages(const GLuint vertex, const GLuint fragment) {
	const GLuint program = glCreateProgram();
	if (program == 0) {
		fprintf(stderr, "ERROR: failed to allocate new shader\n");
		return 0;
	}
	glAttachShader(program, vertex);
	glAttachShader(program, fragment);
	glBindAttribLocation(program, 0, "???position");
	glBindAttribLocation(program, 1, "???uv");
	glLinkProgram(program);
	glDeleteShader(vertex);
	glDeleteShader(fragment);

	GLint succeeded = GL_FALSE;
	glGetProgramiv(program, GL_LINK_STATUS, &succeeded);
	if (succeeded != GL_TRUE) {
		char buffer[1024];
		GLsizei length = 0;
		glGetProgramInfoLog(program, sizeof(buffer), &length, buffer);
		fprintf(stderr, "ERROR: failed to link shader\n%.*s", length, buffer);
		return 0;
	}
	return program;
}
